import _sumBy from 'lodash/sumBy';
import _max from 'lodash/max';
import _range from 'lodash/range';

import { scaleBand, scaleLinear } from 'd3-scale';
import { stack } from 'd3-shape';

/**
 * Generates indexed scale.
 *
 * @param {Array.<Object>} data
 * @param {Function}       getIndex
 * @param {Array.<number>} range
 * @param {number}         padding
 * @returns {Function}
 */
export var getIndexedScale = function getIndexedScale(data, getIndex, range, padding) {
    return scaleBand().rangeRound(range).domain(data.map(getIndex)).padding(padding);
};

/**
 * Generates scale for grouped bar chart.
 *
 * @param {Array.<Object>} data
 * @param {Array.<string>} keys
 * @param {number}         minValue
 * @param {number|string}  _maxValue
 * @param {Array.<number>} range
 * @returns {Function}
 */
export var getGroupedScale = function getGroupedScale(data, keys, minValue, _maxValue, range) {
    var maxValue = _maxValue;
    if (maxValue === 'auto') {
        maxValue = _max(data.reduce(function (acc, entry) {
            return [].concat(acc, keys.map(function (k) {
                return entry[k];
            }));
        }, []));
    }

    return scaleLinear().rangeRound(range).domain([minValue, maxValue]);
};

/**
 * Generates scale for stacked bar chart.
 *
 * @param {Array.<Object>} data
 * @param {Array.<string>} keys
 * @param {number}         minValue
 * @param {number|string}  _maxValue
 * @param {Array.<number>} range
 * @returns {Function}
 */
export var getStackedScale = function getStackedScale(data, keys, minValue, _maxValue, range) {
    var maxValue = _maxValue;
    if (maxValue === 'auto') {
        maxValue = _max(data.map(function (d) {
            return _sumBy(keys, function (key) {
                return d[key];
            });
        }));
    }

    return scaleLinear().rangeRound(range).domain([minValue, maxValue]);
};

/**
 * Generates x/y scales & bars for vertical grouped bar chart.
 *
 * @param {Array.<Object>} data
 * @param {Function}       getIndex
 * @param {Array.<string>} keys
 * @param {number}         minValue
 * @param {number}         maxValue
 * @param {number}         width
 * @param {number}         height
 * @param {Function}       getColor
 * @param {number}         [padding=0]
 * @return {{ xScale: Function, yScale: Function, bars: Array.<Object> }}
 */
export var generateVerticalGroupedBars = function generateVerticalGroupedBars(_ref) {
    var data = _ref.data,
        getIndex = _ref.getIndex,
        keys = _ref.keys,
        minValue = _ref.minValue,
        maxValue = _ref.maxValue,
        width = _ref.width,
        height = _ref.height,
        getColor = _ref.getColor,
        _ref$padding = _ref.padding,
        padding = _ref$padding === undefined ? 0 : _ref$padding;

    var xScale = getIndexedScale(data, getIndex, [0, width], padding);
    var yScale = getGroupedScale(data, keys, minValue, maxValue, [height, 0]);

    var barWidth = xScale.bandwidth() / keys.length;

    var bars = [];
    if (barWidth > 0) {
        keys.forEach(function (key, i) {
            _range(xScale.domain().length).forEach(function (index) {
                var x = xScale(getIndex(data[index])) + barWidth * i;
                var y = yScale(data[index][key]);
                var barHeight = height - y;

                if (barWidth > 0 && barHeight > 0) {
                    var barData = {
                        id: key,
                        value: data[index][key],
                        index: index,
                        indexValue: getIndex(data[index]),
                        data: data[index]
                    };

                    bars.push({
                        key: key + '.' + barData.indexValue,
                        data: barData,
                        x: x,
                        y: y,
                        width: barWidth,
                        height: barHeight,
                        color: getColor(barData)
                    });
                }
            });
        });
    }

    return { xScale: xScale, yScale: yScale, bars: bars };
};

/**
 * Generates x/y scales & bars for horizontal grouped bar chart.
 *
 * @param {Array.<Object>} data
 * @param {Function}       getIndex
 * @param {Array.<string>} keys
 * @param {number}         minValue
 * @param {number}         maxValue
 * @param {number}         width
 * @param {number}         height
 * @param {Function}       getColor
 * @param {number}         [padding=0]
 * @return {{ xScale: Function, yScale: Function, bars: Array.<Object> }}
 */
export var generateHorizontalGroupedBars = function generateHorizontalGroupedBars(_ref2) {
    var data = _ref2.data,
        getIndex = _ref2.getIndex,
        keys = _ref2.keys,
        minValue = _ref2.minValue,
        maxValue = _ref2.maxValue,
        width = _ref2.width,
        height = _ref2.height,
        getColor = _ref2.getColor,
        _ref2$padding = _ref2.padding,
        padding = _ref2$padding === undefined ? 0 : _ref2$padding;

    var xScale = getGroupedScale(data, keys, minValue, maxValue, [0, width]);
    var yScale = getIndexedScale(data, getIndex, [height, 0], padding);

    var barHeight = yScale.bandwidth() / keys.length;

    var bars = [];
    if (barHeight > 0) {
        keys.forEach(function (key, i) {
            _range(yScale.domain().length).forEach(function (index) {
                var x = 0;
                var y = yScale(getIndex(data[index])) + barHeight * i;
                var barWidth = xScale(data[index][key]);

                if (barWidth > 0) {
                    var barData = {
                        id: key,
                        value: data[index][key],
                        index: index,
                        indexValue: getIndex(data[index]),
                        data: data[index]
                    };

                    bars.push({
                        key: key + '.' + barData.indexValue,
                        data: barData,
                        x: x,
                        y: y,
                        width: barWidth,
                        height: barHeight,
                        color: getColor(barData)
                    });
                }
            });
        });
    }

    return { xScale: xScale, yScale: yScale, bars: bars };
};

/**
 * Generates x/y scales & bars for grouped bar chart.
 *
 * @param {Object} options
 * @return {{ xScale: Function, yScale: Function, bars: Array.<Object> }}
 */
export var generateGroupedBars = function generateGroupedBars(options) {
    return options.layout === 'vertical' ? generateVerticalGroupedBars(options) : generateHorizontalGroupedBars(options);
};

/**
 * Generates x/y scales & bars for vertical stacked bar chart.
 *
 * @param {Array.<Object>} data
 * @param {Function}       getIndex
 * @param {Array.<string>} keys
 * @param {number}         minValue
 * @param {number}         maxValue
 * @param {number}         width
 * @param {number}         height
 * @param {Function}       getColor
 * @param {number}         [padding=0]
 * @return {{ xScale: Function, yScale: Function, bars: Array.<Object> }}
 */
export var generateVerticalStackedBars = function generateVerticalStackedBars(_ref3) {
    var data = _ref3.data,
        getIndex = _ref3.getIndex,
        keys = _ref3.keys,
        minValue = _ref3.minValue,
        maxValue = _ref3.maxValue,
        width = _ref3.width,
        height = _ref3.height,
        getColor = _ref3.getColor,
        _ref3$padding = _ref3.padding,
        padding = _ref3$padding === undefined ? 0 : _ref3$padding;

    var xScale = getIndexedScale(data, getIndex, [0, width], padding);
    var yScale = getStackedScale(data, keys, minValue, maxValue, [height, 0]);

    var stackedData = stack().keys(keys)(data);

    var bars = [];
    var barWidth = xScale.bandwidth();

    if (barWidth > 0) {
        stackedData.forEach(function (stackedDataItem) {
            xScale.domain().forEach(function (index, i) {
                var d = stackedDataItem[i];
                var x = xScale(getIndex(d.data));

                var y = yScale(d[1]);
                var barHeight = yScale(d[0]) - y;

                if (barHeight > 0) {
                    var barData = {
                        id: stackedDataItem.key,
                        value: d.data[stackedDataItem.key],
                        index: i,
                        indexValue: index,
                        data: d.data
                    };

                    bars.push({
                        key: stackedDataItem.key + '.' + index,
                        data: barData,
                        x: x,
                        y: y,
                        width: barWidth,
                        height: barHeight,
                        color: getColor(barData)
                    });
                }
            });
        });
    }

    return { xScale: xScale, yScale: yScale, bars: bars };
};

/**
 * Generates x/y scales & bars for horizontal stacked bar chart.
 *
 * @param {Array.<Object>} data
 * @param {Function}       getIndex
 * @param {Array.<string>} keys
 * @param {number}         minValue
 * @param {number}         maxValue
 * @param {number}         width
 * @param {number}         height
 * @param {Function}       getColor
 * @param {number}         [padding=0]
 * @return {{ xScale: Function, yScale: Function, bars: Array.<Object> }}
 */
export var generateHorizontalStackedBars = function generateHorizontalStackedBars(_ref4) {
    var data = _ref4.data,
        getIndex = _ref4.getIndex,
        keys = _ref4.keys,
        minValue = _ref4.minValue,
        maxValue = _ref4.maxValue,
        width = _ref4.width,
        height = _ref4.height,
        getColor = _ref4.getColor,
        _ref4$padding = _ref4.padding,
        padding = _ref4$padding === undefined ? 0 : _ref4$padding;

    var xScale = getStackedScale(data, keys, minValue, maxValue, [0, width]);
    var yScale = getIndexedScale(data, getIndex, [height, 0], padding);

    var stackedData = stack().keys(keys)(data);

    var bars = [];
    var barHeight = yScale.bandwidth();

    if (barHeight > 0) {
        stackedData.forEach(function (stackedDataItem) {
            yScale.domain().forEach(function (index, i) {
                var d = stackedDataItem[i];
                var y = yScale(getIndex(d.data));

                var barData = {
                    id: stackedDataItem.key,
                    value: d.data[stackedDataItem.key],
                    index: i,
                    indexValue: index,
                    data: d.data
                };

                var x = xScale(d[0]);
                var barWidth = xScale(d[1]) - x;

                if (barWidth > 0) {
                    bars.push({
                        key: stackedDataItem.key + '.' + index,
                        data: barData,
                        x: x,
                        y: y,
                        width: barWidth,
                        height: barHeight,
                        color: getColor(barData)
                    });
                }
            });
        });
    }

    return { xScale: xScale, yScale: yScale, bars: bars };
};

/**
 * Generates x/y scales & bars for stacked bar chart.
 *
 * @param {Object} options
 * @return {{ xScale: Function, yScale: Function, bars: Array.<Object> }}
 */
export var generateStackedBars = function generateStackedBars(options) {
    return options.layout === 'vertical' ? generateVerticalStackedBars(options) : generateHorizontalStackedBars(options);
};